#ifndef _SXX_CLIENT_SOCKET_
#define _SXX_CLIENT_SOCKET_

#include <exception>
#include <string>
using std::exception;
using std::string;

namespace sxx_socket {

class client_socket {
private:
	int fd;
	int create_socket();
	void init_socket();

	static bool is_space(char x);

	static void throw_receive_timeout();
	static void throw_transfer_timeout();
	static void throw_creation_exception();
	static void throw_connection_exception();

public:
	client_socket(int fd = -1);
	client_socket(const client_socket& other);
	const client_socket& operator = (const client_socket& other);
	~client_socket();

	void connect_to(const string& host, const string& port);

	char get_char();
	string get_word();
	string get_string(size_t size);
	string get_line();

	void put_char(const char& content);
	void put_word(const string& content);
	void put_string(const string& content);
	void put_line(const string& content);
	void put_line();

	void set_receive_timeout(int seconds);
	void set_transfer_timeout(int seconds);

};

class socket_failure: public exception {

public:
	virtual const char* what() const throw();

};

class socket_receive_timeout: public socket_failure {

public:
	virtual const char* what() const throw();

};

class socket_transfer_timeout: public socket_failure {

public:
	virtual const char* what() const throw();

};

class socket_creation_exception: public socket_failure {

public:
	virtual const char* what() const throw();

};

class socket_connection_exception: public socket_failure {

public:
	virtual const char* what() const throw();

};

} // end of namespace sxx_socket.

#endif

