#ifndef _SXX_CONNECTION_MANAGER_
#define _SXX_CONNECTION_MANAGER_

#include "client_socket.h"
#include <pthread.h>
#include <map>
using namespace sxx_socket;
using std::map;

namespace sxx_manager {

class connection_info {
public:
	client_socket client;
	string host;
	string port;

	connection_info(const client_socket& client,
	                const string& host, const string& port);
	connection_info(const connection_info& value);
};

class connection_manager {
private:
	map<pthread_t, connection_info*> running_threads;
	map<pthread_t, connection_info*> collect_threads;
	pthread_mutex_t running_threads_lock;
	pthread_mutex_t collect_threads_lock;
	size_t thread_limit;

	void (*handler)(connection_info&);
	static void* handler_wrapper(void* argv);
	static void* thread_collector(void* argv);

	connection_manager();

public:
	static connection_manager& get_instance();
	static void signal_handler(int signal);

	void set_handler(void (*handler)(connection_info&));
	void set_limit(size_t count);

	void new_connection(connection_info& info);
	void close_connection(pthread_t thread);
	void close_all_connections();

};

}

#endif

