#ifndef _SXX_FILE_MANAGER_
#define _SXX_FILE_MANAGER_

#include <map>
#include <string>
#include <fstream>
#include <exception>
#include <pthread.h>
#include <time.h>
using namespace std;

namespace sxx_file {

class file_t {
private:
	pthread_mutex_t write_mutex;
	pthread_mutex_t read_mutex;
	string filename;

public:
	file_t();
	file_t(const file_t& other);
	file_t(const string& filename, bool no_read);
	~file_t();

	void write_lock();
	void read_lock();
	void release_write_lock();
	void release_read_lock();

	time_t last_access;
	string content;
};

class file_manager {
private:
	map<string, file_t> cache;
	pthread_mutex_t master_lock;
	file_manager();
	file_t& locate_file(string name, bool no_read);

	static file_manager* me;

public:
	~file_manager();
	static file_manager& instance();
	static void destroy();
	string read_file(const string& filename);
	void write_file(const string& filename, const string& content);
};

class file_manager_failure: public exception {
public:
	virtual const char* what() const throw();
};

class access_denied: public file_manager_failure {
public:
	virtual const char* what() const throw();
};

} // end of namespace sxx_file

#endif

