#include "request_parser.h"

namespace sxx_parser {

bool request_parser::is_space(char x) {
	return isspace(x);
}

void request_parser::chop(string& str) {
	string::iterator itr;
	string::reverse_iterator back_itr;

	for (back_itr = str.rbegin(); back_itr != str.rend(); ++back_itr)
		if (!is_space(*back_itr))
			break;
        str.erase(str.length() - (back_itr - str.rbegin()));

	for (itr = str.begin(); itr != str.end(); ++itr)
		if (!is_space(*itr))
			break;
	str.erase(str.begin(), itr);
}

void request_parser::capitalize(string& str) {
	for (string::iterator i = str.begin(); i != str.end(); ++i) {
		if (*i >= 'a' && *i <= 'z') {
			*i -= 'a';
			*i += 'A';
		}
	}
}

char request_parser::base64_decode_char(char x) {
	if (x >= 'A' && x <= 'Z')
		return x - 'A';
	else if (x >= 'a' && x <= 'z')	
		return x - 'a' + 26;
	else if (x >= '0' && x <= '9')
		return x - '0' + 52;
	else if (x == '+')
		return 62;
	else if (x == '-')
		return 63;
	else
		throw_base64_exception();

	return 0; // just to mute the compiler warning.
}

void request_parser::base64_decode(string& str) {
	int buffer = 0;
	int buffered_bits = 0;
	string::iterator write_itr = str.begin();
	string::iterator read_itr = str.begin();

	if (str.length() % 4)
		throw_base64_exception();

	if (str == "")
		return;

	for (int i = 0; i < 2; ++i) {
		if (str[str.length() - 1] == '=')
			str.erase(str.length() - 1);
		else
			break;
	}

	while (read_itr != str.end()) {
		buffer = buffer << 6 + base64_decode_char(*read_itr);
		buffered_bits += 6;

		while (buffered_bits >= 8) {
			buffered_bits -= 8;
			*write_itr = buffer >> buffered_bits; 
			buffer -= (buffer >> buffered_bits) << buffered_bits;
			++write_itr;
		}

		++read_itr;
	}

	str.erase(write_itr, str.end());
	str.resize(str.length());
}

char request_parser::hex_decode(char x) {
	if (x >= '0' && x <= '9')
		return x - '0';
	else if (x >= 'A' && x <= 'F')
		return x - 'A' + 10;
	else if (x >= 'a' && x <= 'f')
		return x - 'a' + 10;
	else
		throw_url_exception();

	return 0; // just to mute the compiler warning.
}

void request_parser::url_decode(string& str) {
	string::iterator write_itr = str.begin();
	string::iterator read_itr = str.begin();

	while (read_itr != str.end()) {
		// If it is not a special character, go ahead.
		if (*read_itr == '+') {
			++read_itr;
			*write_itr = ' ';
			++write_itr;
			continue;
		} else if (*read_itr != '%') {
			*write_itr = *read_itr;
			++read_itr;
			++write_itr;
			continue;
		}

		// At this point, we are going to decode a character from hex.
		// Check to make sure we are not too close to end of string.
		if (str.end() - read_itr < 3)
			throw_parameter_exception();

		*write_itr = hex_decode(*(read_itr + 1)) * 16
		             + hex_decode(*(read_itr + 2));

		read_itr += 3;
		++write_itr;
	}

	str.erase(write_itr, str.end());
	str.resize(str.length());
}

vector<string> request_parser::split(const string& str, char delim, int n) {
	vector<string::const_iterator> delim_list;
	string::const_iterator last_delim;
	vector<string> result;

	// Get the list of delimiters.
	for (string::const_iterator i = str.begin(); i != str.end(); ++i) {
		if (*i == delim) {
			delim_list.push_back(i);
			if (--n == 1)
				break;
		}
	}
	delim_list.push_back(str.end());

	// With the list of delimiters, dig out the contents.
	last_delim = str.begin();
	for (size_t i = 0; i < delim_list.size(); ++i) {
		if (delim_list[i] != last_delim)
			result.push_back(string(last_delim, delim_list[i]));
		last_delim = delim_list[i] + 1;
	}

	return result;
}                       

var& request_parser::locate_var(var& info, const string& str) {
	string index;
	string next_index;
	string::const_iterator itr;

	if (str == "")
		return info;

	for (itr = str.begin(); itr != str.end(); ++itr) {
		if (*itr == '[') {
			if (*str.begin() == '[' && itr != str.begin()) {
				// [abc[...
				throw_parameter_exception();
			} else if (itr != str.begin()) {
				// abc[...
				break;
			} else {
				// [...
				continue;
			}
		}

		if (*itr == ']') {
			if (*str.begin() == '[') {
				// [abc]...
				++itr;
				break;
			} else {
				// abc]...
				throw_parameter_exception();
			}
		}

		if (*itr != '[')
			index += *itr;
	}

	next_index = string(itr, str.end());

	try {
		if (index == "") // new element to a vector.
			return locate_var(info[info.size()], next_index);
		else // new element of a map.
			return locate_var(info[index], next_index);
	} catch (var_failure) { }

	throw_parameter_exception();
	return info; // never gets to here; just to mute warning.
}

void request_parser::throw_base64_exception() {
	request_parser_base64_exception e;
	throw e;
}

void request_parser::throw_url_exception() {
	request_parser_url_exception e;
	throw e;
}

void request_parser::throw_initial_exception() {
	request_parser_initial_exception e;
	throw e;
}

void request_parser::throw_parameter_exception() {
	request_parser_parameter_exception e;
	throw e;
}

void request_parser::throw_length_exception() {
	request_parser_length_exception e;
	throw e;
}

void request_parser::initial(request& info) {
	string line;

	do {
		line = input.get_line();
	} while (not_first && line == "");
	vector<string> elements = split(line, ' ');

	if (elements.size() != 3)
		throw_initial_exception();

	method(info, elements[0]);
	if (elements[1][0] == '/')
		path(info, elements[1]);
	else
		url(info, elements[1]);
	protocol(info, elements[2]);

	not_first = true;
}

void request_parser::method(request& info, string& str) {
	capitalize(str);
	info.method = str;
}

void request_parser::path(request& info, string& str) {
	vector<string> parts = split(str, '?', 2);

	url_decode(parts[0]);
	info.path = parts[0];

	if (parts.size() == 2)
		parameters(info.get, parts[1], '&');
}

void request_parser::url(request& info, string& str) {
	string::iterator start_itr = str.begin();
	int skipped_slashes = 0;

	while (start_itr != str.end()) {
		if (*start_itr == '/') {
			if (++skipped_slashes == 3)
				break;
		}

		++start_itr;
	}

	string path_str(start_itr, str.end());
	path(info, path_str);
}

void request_parser::protocol(request& info, string& str) {
	info.protocol = str;
}

void request_parser::parameters(var& info, const string& str, char delim) {
	vector<string> pairs = split(str, delim);
	vector<string>::const_iterator itr;

	for (itr = pairs.begin(); itr != pairs.end(); ++itr) {
		size_t pos;
		string key;
		string value;

		if (itr->empty())
			continue;

		pos = itr->find("=");

		if (pos != string::npos) {
			key = itr->substr(0, pos);
			value = itr->substr(pos + 1);
		} else {
			key = *itr;
			// value = "";
		}

		chop(key);
		chop(value);
		url_decode(key);
		url_decode(value);
		locate_var(info, key) = value;
	}
}

void request_parser::header(request& info) {
	while (true) {
		vector<string> parts;
		string line = input.get_line();

		if (line == "")
			break;

		// If we get space as the first character, it is considereda as
		// continuation of the previous header line.
		if (!is_space(line[0])) {
			parts = split(line, ':', 2);
			capitalize(parts[0]);
			chop(parts[0]);
			chop(parts[1]);
		} else {
			chop(line);
			parts.push_back(last_header);
			parts.push_back(info.headers[last_header] + " " + line);
		}

		if (parts[0] == "HOST")
			host(info, parts[1]);
		else if (parts[0] == "COOKIE")
			cookie(info, parts[1]);

		info.headers[parts[0]] = parts[1];
		last_header = parts[0];
	}
}

void request_parser::host(request& info, string& str) {
	url_decode(str);
	info.host = str;
}

void request_parser::cookie(request& info, string& str) {
	parameters(info.cookie, str, ';');
}

void request_parser::post(request& info) {
	if (!(info.method == "POST" || info.method == "PUT"))
		return;

	if (info.headers["CONTENT-TYPE"].find("multipart") != string::npos)
		multipart_post(info);
	else
		single_post(info);
}

void request_parser::single_post(request& info) {
	try {
		size_t size = (var)info.headers["CONTENT-LENGTH"];
		string post = input.get_string(size);

		if (info.headers["CONTENT-TYPE"]
		    == "application/x-www-form-urlencoded")
			parameters(info.post, post, '&');
		else
			info.post = post;
	} catch (const var_failure& e) {
		throw_length_exception();
	}

}

void request_parser::multipart_post(request& info) {
	throw_length_exception();
}

request_parser::request_parser(const client_socket& input) {
	this->input = input;
	not_first = false;
}

request request_parser::get_request() {
	request info;

	initial(info);
	header(info);
	post(info);

	return info;
}

const char* request_parser_failure::what() const throw() {
	return "Unable to parse request.";
}

const char* request_parser_base64_exception::what() const throw() {
	return "Invalid base64 encoding.";
}

const char* request_parser_url_exception::what() const throw() {
	return "Invalid url encoding.";
}

const char* request_parser_initial_exception::what() const throw() {
	return "Invalid request line.";
}

const char* request_parser_parameter_exception::what() const throw() {
	return "Invalid encoded parameters.";
}

const char* request_parser_length_exception::what() const throw() {
	return "Length of attachments required.";
}

} // end of namespace sxx_parser.

