#ifndef _SXX_REQUEST_PARSER_
#define _SXX_REQUEST_PARSER_

#include "client_socket.h"
#include "request.h"
using namespace sxx_socket;
using namespace sxx_request;

namespace sxx_parser {

class request_parser {
private:
	client_socket input;
	bool not_first;
	string last_header;

	static bool is_space(char x);
	static void chop(string& str);
	static void capitalize(string& str);
	static char base64_decode_char(char x);
	static void base64_decode(string& str);
	static char hex_decode(char x);
	static void url_decode(string& str);
	static vector<string> split(const string& str, char delim, int n = -1);
	static var& locate_var(var& info, const string& str);

	static void throw_base64_exception();
	static void throw_url_exception();
	static void throw_initial_exception();
	static void throw_parameter_exception();
	static void throw_length_exception();

	void initial(request& info);
	void method(request& info, string& str);
	void url(request& info, string& str);
	void path(request& info, string& str);
	void parameters(request& info, string& str);
	void protocol(request& info, string& str);
	void parameters(var& info, const string& str, char delim);

	void header(request& info);
	void host(request& info, string& str);
	void cookie(request& info, string& str);

	void post(request& info);
	void single_post(request& info);
	void multipart_post(request& info);

public:
	request_parser(const client_socket& input);
	request get_request();

};

class request_parser_failure: public exception {
public:
	virtual const char* what() const throw();

};

class request_parser_base64_exception: public request_parser_failure {
public:
	virtual const char* what() const throw();

};


class request_parser_url_exception: public request_parser_failure {
public:
	virtual const char* what() const throw();

};

class request_parser_initial_exception: public request_parser_failure {
public:
	virtual const char* what() const throw();

};

class request_parser_parameter_exception: public request_parser_failure {
public:
	virtual const char* what() const throw();

};

class request_parser_length_exception: public request_parser_failure {
public:
	virtual const char* what() const throw();

};

} // end of namespace sxx_parser.

#endif

