#ifndef _SXX_SERVER_SOCKET_
#define _SXX_SERVER_SOCKET_

#include "client_socket.h"
#include <netinet/in.h>
using sxx_socket::socket_failure;

namespace sxx_socket {

class server_socket {
private:
	int fd;
	int receive_timeout;
	int transfer_timeout;
	
	int create_socket();
	void init_socket();
	void bind_socket(const struct sockaddr_in* address);
	void listen_socket(int backlog);

	static void throw_accept_timeout();
	static void throw_binding_exception();
	static void throw_listen_exception();
	static void throw_creation_exception();

	server_socket& operator = (const server_socket& other);

public:
	server_socket(const string& host, const string& port, int backlog);

	client_socket accept_client();
	client_socket accept_client(string &host, string& port);
	void set_receive_timeout(int seconds);
	void set_transfer_timeout(int seconds);
	void set_accept_timeout(int seconds);

};

class socket_accept_timeout: public socket_failure {

public:
	virtual const char* what() const throw();

};

class socket_binding_exception: public socket_failure {

public:
	virtual const char* what() const throw();

};

class socket_listen_exception: public socket_failure {

public:
	virtual const char* what() const throw();

};

} // end of namespace sxx_socket.

#endif

