#include "session_manager.h"

typedef std::string str;
typedef sxx_var::var var;
typedef std::pair<time_t, var> ptv;
typedef std::map<str, ptv> s2tv;
typedef sxx_session::session_exception s_ex;

namespace sxx_session {

session_manager* session_manager::me;

const char* session_exception::what() const throw() {
	return "Invalid or expired session.";
}

pthread_t session_manager::recycle_thread;
pthread_mutex_t session_manager::lock;
s2tv session_manager::table;

session_manager::session_manager() {
	pthread_mutex_init(&lock, NULL);
	pthread_create(&recycle_thread, NULL, session_recycle, NULL);
}

session_manager& session_manager::instance() {
	static bool init = false;

	if (!init) {
		me = new session_manager();
		init = true;
	}

	return *me;
}

void session_manager::destroy() {
	delete me;
	me = 0;
}

void* session_manager::session_recycle(void* args) {
	s2tv::iterator i;

	while (true) {
		time_t water_mark = time(NULL) - 3600;
		pthread_mutex_lock(&lock);

		for (i = table.begin(); i != table.end(); ) 
			if (i->second.first < water_mark)
				table.erase(i);
			else
				++i;

		pthread_mutex_unlock(&lock);
		sleep(1);
	}
}

var& session_manager::session_var(str id) {
	s2tv::iterator record;

	record = table.find(id);
	if (record != table.end())
		return record->second.second;

	throw s_ex();
}

str session_manager::new_session() {
	static unsigned long long rand1; // = some garbage.
	static unsigned long long rand2; // = some garbage.
	static unsigned long long rand[] = { 2, 3, 5, 7, 11, 29 };
	time_t current_time;
	str new_key;

	current_time = time(NULL);
	rand1 += rand[rand2 % (sizeof(rand) / sizeof(rand1))];
	rand2 += rand[rand1 % (sizeof(rand) / sizeof(rand2))];

	new_key = (string)(var)current_time;
	new_key += (string)(var)rand1;
	new_key += (string)(var)rand2;

	table[new_key] = ptv(current_time, var());

	return new_key;
}

void session_manager::close_session(str id) {
	s2tv::iterator to_erase;

	pthread_mutex_lock(&lock);
	
	to_erase = table.find(id);
	if (to_erase != table.end())
		table.erase(to_erase);

	pthread_mutex_unlock(&lock);
}

}

