#include "sxx_main.h"
#include <iostream>
using namespace std;

string sxx_hostname = "localhost";
string sxx_port = "8080";
int sxx_backlog = 10;
int sxx_max_threads = 100;
int sxx_receive_timeout = 10;
int sxx_transfer_timeout = 60;

int main(int argc, char* argv[]) { try {
	cout << "Initializing... " << flush;

	sxx_config();
	sxx_init();

	server_socket server(sxx_hostname, sxx_port, sxx_backlog);
	server.set_receive_timeout(sxx_receive_timeout);
	server.set_transfer_timeout(sxx_transfer_timeout);

	connection_manager& manager = connection_manager::get_instance();
	manager.set_handler(sxx_client_handler);
	manager.set_limit(sxx_max_threads);

	signal(SIGTERM, sxx_shutdown_handler);
	signal(SIGINT, sxx_shutdown_handler);
	signal(SIGQUIT, sxx_shutdown_handler);
	signal(SIGHUP, connection_manager::signal_handler);
	signal(SIGABRT, connection_manager::signal_handler);
	signal(SIGSEGV, connection_manager::signal_handler);
	signal(SIGBUS, connection_manager::signal_handler);
	signal(SIGPIPE, connection_manager::signal_handler);
	
	cout << "Now ready." << endl;

	while (true) { try {
		string host;
		string port;
		client_socket client = server.accept_client(host, port);
		connection_info info(client, host, port);

		manager.new_connection(info);
	} catch (const exception& e) { /* cerr << e.what() << endl; */ } }
} catch (const exception& e) { /* cerr << e.what() << endl; */ } }

void sxx_client_handler(connection_info& info) {
	request_parser parser(info.client);
	response_generator generator(info.client);

	while (true) { try {
			request input = parser.get_request();
			response output;
			string sid = (string)input.cookie["_sxx_sid"];
			var page_info;

			input.remote_host = info.host;
			input.remote_port = info.port;
			try {
				sxx_handler(input,
				            output,
					    page_info,
					    session_manager::instance().session_var(sid));
			} catch (const session_exception& err) {
				sid = session_manager::instance().new_session();
				sxx_handler(input,
				            output,
					    page_info,
					    session_manager::instance().session_var(sid));
				output.session_cookie["_sxx_sid"] = sid;
			}
			generator.put_response(output);
		} catch (const socket_failure& e) {
			// "Do nothing" if the client is disconnected.
			// Socket closes once the class is out of scope.
			break;
		} catch (const exception& e) { try {
			// Try to report other errors.
			response output;
			output.status = 500;
			output.connection_close = true;
			output.content << e.what() << endl;
			generator.put_response(output);
			break;
		} catch (const exception& err) { break; } }
	}
}

void sxx_null_handler(int signal) {
	// cerr << "Nasty signal " << signal << " caught." << endl;
}

void sxx_shutdown_handler(int signal) {
	cout << "Shutting down... " << flush;
	connection_manager::get_instance().close_all_connections();
	sxx_destroy();
	file_manager::destroy();
	session_manager::destroy();
	cout << "Finished." << endl;
	_exit(0);
}
