%{

#include <iostream>
#include <vector>
#include <string>
using namespace std;

string input_file = "input";
string output_file = "output";
string function_name;

typedef int source;
typedef int restriction;

source GET = 1;
source POST = 2;
source ECOOKIE = 3;
source SCOOKIE = 4;
source PCOOKIE = 5;
source PAGE = 6;
source SESSION = 7;
source READ = 8;
source WRITE = 9;

restriction MINV = 1;
restriction MAXV = 2;
restriction MINL = 3;
restriction MAXL = 4;
restriction REGEX = 5;
restriction MATCH = 6;

struct parameter_t {
	int line_count;
	vector<string> def;
	vector<string> init;
	string name;
	vector<source> src;
	vector<string> src_str;
	vector<restriction> res;
	vector<string> res_str;
	string err_str;
};

vector<string> parameter_buffer;
vector<parameter_t> parameters;

int para_count = 0;
int input_line = 1;
int output_line = 1;

void error(const string& error);
void begin_header();
void end_header();
void begin_parameter();
void end_parameter();
void begin_parameter_init();
void end_parameter_init();
void begin_parameter_modifier();
void end_parameter_modifier();
void begin_body();
void end_body();

%}


%x header
%x parameter
%x parameter_modifier
%x modifier_content
%x restriction_content
%x modifier_content_str
%x restriction_content_str
%x error_content
%x error_content_str
%x parameter_init
%x parameter_init_str
%x body
%x body_comment
%x function_body
%x function_body_comment

%%

<header>\n                      ECHO; ++input_line; ++output_line;
<header>[^%\n]*                 ECHO;
<header>^%%$                    end_header(); begin_parameter();
<header><<EOF>>                 error("unexpected end-of-file");

<parameter>[ \t\r]*             ;
<parameter>"//"+[^\n]*          ;
<parameter>\n                   ++input_line;
<parameter>"const"              error("const variable not allowed");
<parameter>"="                  end_parameter(); begin_parameter_init();
<parameter>";"                  end_parameter(); begin_parameter_modifier();
<parameter>[^ \t\r\n;=]*        parameter_buffer.push_back(yytext);
<parameter>^%%$                 begin_body();
<parameter><<EOF>>              error("unexpected end-of-file");

<parameter_init>"//"+[^\n]*     ;
<parameter_init>\n              ++input_line; parameters.back().init.back() += " ";
<parameter_init>";"             end_parameter_init(); begin_parameter_modifier();
<parameter_init>\"              BEGIN(parameter_init_str); parameters.back().init.back() += yytext;
<parameter_init>.               parameters.back().init.back() += yytext;
<parameter_init><<EOF>>         error("unexpected end-of-file");

<parameter_init_str>\"          parameters.back().init.back() += yytext; BEGIN(parameter_init);
<parameter_init_str>\\\"        parameters.back().init.back() += yytext;
<parameter_init_str>[^\n]       parameters.back().init.back() += yytext;
<parameter_init_str>\n          error("unbounded string literal");
<parameter_init_str><<EOF>>     error("unexpected end-of-file");

<modifier_content_str>\"        parameters.back().src_str.back() += yytext; BEGIN(modifier_content);
<modifier_content_str>\\\"      parameters.back().src_str.back() += yytext;
<modifier_content_str>[^\n]     parameters.back().src_str.back() += yytext;
<modifier_content_str>\n        error("unbounded string literal");
<modifier_content_str><<EOF>>   error("unexpected end-of-file");

<restriction_content_str>\"     parameters.back().res_str.back() += yytext; BEGIN(restriction_content);
<restriction_content_str>\\\"   parameters.back().res_str.back() += yytext;
<restriction_content_str>[^\n]  parameters.back().res_str.back() += yytext;
<restriction_content_str>\n     error("unbounded string literal");
<restriction_content_str><<EOF>> error("unexpected end-of-file");

<error_content_str>\"           parameters.back().err_str += yytext; BEGIN(error_content);
<error_content_str>\\\"         parameters.back().err_str += yytext;
<error_content_str>[^\n]        parameters.back().err_str += yytext;
<error_content_str>\n           error("unbounded string literal");

<parameter_modifier>"//"+[^\n]* ;
<parameter_modifier>[ \t\r]*    ;
<parameter_modifier>\\(\r)\n    ++input_line;
<parameter_modifier>\n          if (parameters.back().src.empty()) error("parameter without source"); end_parameter_modifier(); ++input_line;
<parameter_modifier>"GET"       parameters[parameters.size() - 1].src.push_back(GET); parameters.back().src_str.push_back("");
<parameter_modifier>"POST"      parameters[parameters.size() - 1].src.push_back(POST); parameters.back().src_str.push_back("");
<parameter_modifier>"COOKIE"    parameters[parameters.size() - 1].src.push_back(SCOOKIE); parameters.back().src_str.push_back("");
<parameter_modifier>"ECOOKIE"   parameters[parameters.size() - 1].src.push_back(ECOOKIE); parameters.back().src_str.push_back("");
<parameter_modifier>"SCOOKIE"   parameters[parameters.size() - 1].src.push_back(SCOOKIE); parameters.back().src_str.push_back("");
<parameter_modifier>"PCOOKIE"   parameters[parameters.size() - 1].src.push_back(PCOOKIE); parameters.back().src_str.push_back("");
<parameter_modifier>"PAGE"      parameters[parameters.size() - 1].src.push_back(PAGE); parameters.back().src_str.push_back("");
<parameter_modifier>"SESSION"   parameters[parameters.size() - 1].src.push_back(SESSION); parameters.back().src_str.push_back("");
<parameter_modifier>"READ("     parameters.back().src.push_back(READ); parameters.back().src_str.push_back(""); BEGIN(modifier_content);
<parameter_modifier>"WRITE("    parameters.back().src.push_back(WRITE); parameters.back().src_str.push_back(""); BEGIN(modifier_content);
<parameter_modifier>"MINV("     parameters.back().res.push_back(MINV); parameters.back().res_str.push_back(""); BEGIN(restriction_content);
<parameter_modifier>"MAXV("     parameters.back().res.push_back(MAXV); parameters.back().res_str.push_back(""); BEGIN(restriction_content);
<parameter_modifier>"MINL("     parameters.back().res.push_back(MINL); parameters.back().res_str.push_back(""); BEGIN(restriction_content);
<parameter_modifier>"MAXL("     parameters.back().res.push_back(MAXL); parameters.back().res_str.push_back(""); BEGIN(restriction_content);
<parameter_modifier>"REGEX("    parameters.back().res.push_back(REGEX); parameters.back().res_str.push_back(""); BEGIN(restriction_content);
<parameter_modifier>"MATCH("    parameters.back().res.push_back(MATCH); parameters.back().res_str.push_back(""); BEGIN(restriction_content);
<parameter_modifier>"ERROR("    parameters.back().err_str = ""; BEGIN(error_content);
<parameter_modifier>.           error("unknown modifier");
<parameter_modifier><<EOF>>     error("unexpected end-of-file");

<modifier_content>"//"          error("comment not allowed");
<modifier_content>"/*"          error("comment_not_allowed");
<modifier_content>"("           ++para_count; parameters.back().src_str.back() += yytext;
<modifier_content>")"           --para_count; if (para_count < 0) { para_count = 0; BEGIN(parameter_modifier); } else { parameters.back().src_str.back() += yytext; }
<modifier_content>\"            parameters.back().src_str.back() += yytext; BEGIN(modifier_content_str);
<modifier_content>[^\n()\"]*    parameters.back().src_str.back() += yytext;
<modifier_content>\n            error("unexpected end-of-line in modifier");
<modifier_content><<EOF>>       error("unexpected end-of-file");

<restriction_content>"//"       error("comment not allowed");
<restriction_content>"/*"       error("comment_not_allowed");
<restriction_content>"("        ++para_count; parameters.back().res_str.back() += yytext;
<restriction_content>")"        --para_count; if (para_count < 0) { para_count = 0; BEGIN(parameter_modifier); } else { parameters.back().res_str.back() += yytext; }
<restriction_content>\"         parameters.back().res_str.back() += yytext; BEGIN(restriction_content_str);
<restriction_content>[^\n()\"]* parameters.back().res_str.back() += yytext;
<restriction_content>\n         error("unexpected end-of-line in restriction");
<restriction_content><<EOF>>    error("unexpected end-of-file");

<error_content>"//"             error("comment not allowed");
<error_content>"/*"             error("comment_not_allowed");
<error_content>"("              ++para_count; parameters.back().err_str += yytext;
<error_content>")"              --para_count; if (para_count < 0) { para_count = 0; BEGIN(parameter_modifier); } else { parameters.back().err_str += yytext; }
<error_content>\"               parameters.back().err_str += yytext; BEGIN(error_content_str);
<error_content>[^\n()\"]*       parameters.back().err_str += yytext;
<error_content>\n               error("unexpected end-of-line in restriction");

<body>\n                        ECHO; ++input_line; ++output_line;
<body>[^\n]*                    ECHO;
<body><<EOF>>                   end_body();

%%

void error(const string& str) {
	cerr << input_file << ": error in line " << input_line << ": " << str << endl;
	exit(1);
}

void begin_header() {
	BEGIN(header);

	++output_line;
	cout << "#include \"../../sxx/var.h\"" << endl;
	++output_line;
	cout << "#include \"../../sxx/validate.h\"" << endl;
	++output_line;
	cout << "#include \"../../sxx/request.h\"" << endl;
	++output_line;
	cout << "#include \"../../sxx/response.h\"" << endl;
	++output_line;
	cout << "#include \"../../sxx/file_manager.h\"" << endl;
	++output_line;
	cout << "using namespace sxx_var;" << endl;
	++output_line;
	cout << "using namespace sxx_validate;" << endl;
	++output_line;
	cout << "using namespace sxx_request;" << endl;
	++output_line;
	cout << "using namespace sxx_response;" << endl;
	++output_line;
	cout << "using namespace sxx_file;" << endl;
	++output_line;
	cout << "#define sxx_forward(x) { _sxx_forward = x; goto _sxx_finish; }" << endl;
	++output_line;
	cout << "#define sxx_redirect(x) { output.status = 303; output.location = x; goto _sxx_finish; }" << endl;
	++output_line;
	cout << "#define sxx_read(x) file_manager::instance().read_file(string(\"data/\") + x)" << endl;
	++output_line;
	cout << "#define sxx_write(x, y) file_manager::instance().write_file(string(\"data/\") + x, y)" << endl;
	++output_line;
	cout << endl;

	cout << "#line " << input_line << " \"" << input_file << "\"" << endl;
	++output_line;
}

void end_header() {

}

void begin_parameter() {
	BEGIN(parameter);
}

void end_parameter() {
	parameter_t temp;
	temp.line_count = input_line;
	temp.name = parameter_buffer.back();
	temp.err_str = "\"" + temp.name + "\"";
	parameter_buffer.pop_back();
	temp.def = parameter_buffer;
	parameters.push_back(temp);
	parameter_buffer.clear();
}

void begin_parameter_init() {
	BEGIN(parameter_init);
	parameters.back().init.push_back("");
}

void end_parameter_init() {

}

void begin_parameter_modifier() {
	BEGIN(parameter_modifier);
	para_count = 0;
}

void end_parameter_modifier() {
	BEGIN(parameter);
}

void begin_body() {
	BEGIN(body);
	++output_line;
	cout << "#line " << output_line << " \"" << output_file << "\"" << endl;
	++output_line;
	cout << "string " << function_name << "(const request& input, response& output, var& page, var& session) {" << endl;
	++output_line;
	cout << "std::string _sxx_forward;" << endl;

	for (unsigned i = 0; i < parameters.size(); ++i) {
		++output_line;
		++output_line;
		cout << "#line " << parameters[i].line_count << " \"" << input_file << "\"" << endl;

		for (unsigned j = 0; j < parameters[i].def.size(); ++j) {
			cout << parameters[i].def[j] << " ";
		}
		cout << parameters[i].name << ";" << endl;
	}

	// Print the variables.
	for (unsigned i = 0; i < parameters.size(); ++i) {
		++output_line;
		cout << "#line " << output_line << " \"" << output_file << "\"" << endl;
		++output_line;
		cout << "try { throw 1; }" << endl;

		// Try to name some source of information.
		for (unsigned j = 0; j < parameters[i].src.size(); ++j) {
			++output_line;
			++output_line;
			cout << "// #line " << parameters[i].line_count << " \"" << input_file << "\"" << endl;
			if (parameters[i].src[j] == READ) {
				cout << "catch (...) { try { " << parameters[i].name
				     << " = file_manager::instance().read_file(string(\"data/\") + "
				     << parameters[i].src_str[j]
				     << "); ";
			} else if (parameters[i].src[j] == WRITE) {
				cout << endl;
				continue;
			} else {
				cout << "catch (...) { try { " << parameters[i].name << " = (" << parameters[i].def.back() << ")"
				     << (parameters[i].src[j] == GET? "input.get":
				        parameters[i].src[j] == POST? "input.post":
					parameters[i].src[j] == ECOOKIE? "input.cookie":
					parameters[i].src[j] == SCOOKIE? "input.cookie":
					parameters[i].src[j] == PCOOKIE? "input.cookie":
					parameters[i].src[j] == PAGE? "page":
					parameters[i].src[j] == SESSION? "session": "")
				     << "[\"" << parameters[i].name << "\"]; ";
				cout << "string _sxx_last_var = (string)"
				     << (parameters[i].src[j] == GET? "input.get":
				        parameters[i].src[j] == POST? "input.post":
					parameters[i].src[j] == ECOOKIE? "input.cookie":
					parameters[i].src[j] == SCOOKIE? "input.cookie":
					parameters[i].src[j] == PCOOKIE? "input.cookie":
					parameters[i].src[j] == PAGE? "PAGE":
					parameters[i].src[j] == SESSION? "session": "")
				     << "[\"" << parameters[i].name << "\"]; ";
			}

			// For each naming, check if it is valid.
			// If not, throw exception.
			cout << "if ( 0";
			for (unsigned k = 0; k < parameters[i].res.size(); ++k) {
				if (parameters[i].res[k] == MINV)
					cout << " || " << parameters[i].name << " < " << parameters[i].res_str[k];
				else if (parameters[i].res[k] == MAXV)
					cout << " || " << parameters[i].name << " > " << parameters[i].res_str[k];
				else if (parameters[i].res[k] == MINL)
					cout << " || _sxx_last_var.length() < " << parameters[i].res_str[k];
				else if (parameters[i].res[k] == MAXL)
					cout << " || _sxx_last_var.length() > " << parameters[i].res_str[k];
				else if (parameters[i].res[k] == REGEX)
					cout << " || (regex(_sxx_last_var, " << parameters[i].res_str[k] << ")).size() < 0";
			}
			cout << ") throw 2;";
			cout << " }" << endl;
		}

		// If there is any default value, use them as the backup.
		// Otherwise, throw exception.
		if (parameters[i].init.size()) {
			++output_line;
			++output_line;
			cout << "// #line " << parameters[i].line_count << " \"" << input_file << "\"" << endl;
			cout << "catch (...) { " << parameters[i].name << " =";
			for (unsigned j = 0; j < parameters[i].init.size(); ++j)
				cout << " " << parameters[i].init[j];
			cout << "; }" << endl;
		} else {
			++output_line;
			++output_line;
			cout << "// #line " << parameters[i].line_count << " \"" << input_file << "\"" << endl;
			cout << "catch (...) { throw(" << parameters[i].err_str << "); }" << endl;
		}

		++output_line;
		++output_line;
		cout << "#line " << output_line << " \"" << output_file << "\"" << endl;
		for (unsigned j = 0; j < parameters[i].src.size(); ++j) {
			if (parameters[i].src[j] != WRITE)
				cout << "} ";
		}
		cout << endl;

		// Check that the value is correct, if MATCH() is given.
		for (unsigned j = 0; j < parameters[i].res.size(); ++j) {
			if (parameters[i].res[j] == MATCH) {
				++output_line;
				++output_line;
				cout << "#line " << output_line << " \"" << output_file << "\"" << endl;
				cout << "if (" << parameters[i].name << " != " << parameters[i].res_str[j] << ") throw " << parameters[i].err_str << ";" << endl;
			}
		}
	}

	cout << "#line " << input_line + 1 << " \"" << input_file << "\"" << endl;
}

void end_body() {
	++output_line;
	++output_line;
	cout << "#line " << output_line << " \"" << output_file << "\"" << endl;
	++output_line;
	cout << "goto _sxx_finish;" << endl;
	cout << "_sxx_finish: " << endl;
	for (unsigned i = 0; i < parameters.size(); ++i) {
		for (unsigned j = 0; j < parameters[i].src.size(); ++j) {
			if (parameters[i].src[j] != ECOOKIE &&
			    parameters[i].src[j] != SCOOKIE &&
			    parameters[i].src[j] != PCOOKIE &&
			    parameters[i].src[j] != PAGE &&
			    parameters[i].src[j] != SESSION &&
			    parameters[i].src[j] != WRITE)
			    	continue;

			++output_line;
			++output_line;
			cout << "#line " << parameters[i].line_count << " \"" << input_file << "\"" << endl;
			
			if (parameters[i].src[j] == WRITE) {
				cout << "file_manager::instance().write_file(string(\"data/\") + " << parameters[i].src_str[j] << ", " << parameters[i].name << ");" << endl;
			} else {
				cout << (parameters[i].src[j] == ECOOKIE? "output.expire_cookie":
				         parameters[i].src[j] == SCOOKIE? "output.session_cookie":
				         parameters[i].src[j] == PCOOKIE? "output.persistent_cookie":
				         parameters[i].src[j] == SESSION? "session":
				         parameters[i].src[j] == PAGE? "page": "")
				     << "[\"" << parameters[i].name << "\"]"
				     << " = " << parameters[i].name << ";" << endl;
			}
		}
	}

	++output_line;
	++output_line;
	cout << "#line " << output_line << " \"" << output_file << "\"" << endl;
	cout << "return _sxx_forward;" << endl;
	++output_line;
	cout << "}" << endl;
	exit(0);
}

int main(int argc, char* argv[]) {
	if (argc != 2) {
		cerr << "Usage: " << argv[0] << " function_name" << endl;
		exit(1);
	}

	input_file = argv[1] + string(".page");
	output_file = string("page_") + argv[1] + string(".cc");
	function_name = string("page_") + argv[1];

	begin_header();
	yylex();
}
