#include "validate.h"
#include <regex.h>

namespace sxx_validate {

vector<string> regex(const string& text, const string& pattern) {
	regex_t compiled;
	regmatch_t* matched;
	vector<string> result;
	size_t result_size;

	if (regcomp(&compiled, pattern.c_str(), REG_EXTENDED)) {
		regex_syntax err;
		throw (err);
	}

	result_size = compiled.re_nsub + 1;
	
	if (!(matched = new regmatch_t[result_size])) {
		regex_failure err;
		throw (err);
	}
	
	if (regexec(&compiled, text.c_str(), result_size, matched, 0)) {
		regex_mismatch err;
		throw (err);
	}
	
	for (unsigned i = 0; i < result_size; ++i)
		if (matched[i].rm_so >= 0 &&
		    matched[i].rm_eo >= 0 &&
		    matched[i].rm_so < pattern.length() &&
		    matched[i].rm_eo < pattern.length())
			result.push_back(text.substr(matched[i].rm_so,
			                             matched[i].rm_eo
						     - matched[i].rm_so));
		else
			result.push_back("");
	
	regfree(&compiled);
	delete [] matched;

	return result;
}

const char* regex_failure::what() const throw() {
	return "Regular expression failure.";
}

const char* regex_syntax::what() const throw() {
	return "Regular expression syntax error.";
}

const char* regex_mismatch::what() const throw() {
	return "Pattern mismatch.";
}

}
